/* $Id: codegen.c,v 1.8 2006/01/17 17:21:36 arved Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include "globals.h"
#include "codegen.h"
#include "code.h"

/* For OP_ defines */
#include "y.tab.h"

/* For printnode */
#include "syntree.h"

extern int lookup(char *sym_name);

/* tmpOffset is the memory offset for temps
   It is decremented each time a temp is stored,
   and imcremented when loaded again */
static int tmpOffset = 0;


void trav(TreeNode *tree);

void gen(TreeNode *tree){
	int savedLoc1, savedLoc2, currentLoc;
	int loc;

	switch(tree->kind) {
		case KIND_IF:
			/* savedLoc1 = jump to else
			   savedLoc2 = jump to end */
			emitComment("BEGIN IF");
			/* generate code for expression */
			trav(tree->child[0]);
			/* Save line, to be filled in later */
			savedLoc1 = emitSkip(1);
			emitComment("if: jump to else belongs here");
			/* recurse on "then" part */
			trav(tree->child[1]);
			savedLoc2 = emitSkip(1);
			emitComment("if: jump to end belongs here");
			currentLoc = emitSkip(0);
			emitBackup(savedLoc1);
			emitRM_Abs("JEQ",ac,currentLoc,"if: jmp to else");
			emitRestore();
			/* recurse on else part */
			trav(tree->child[2]);
			currentLoc = emitSkip(0);
			emitBackup(savedLoc2);
			emitRM_Abs("LDA",pc,currentLoc,"jmp to end");
			emitRestore();
			emitComment("END IF");
			break;
		case KIND_WHILE:
			/* savedLoc1 = jump to begin of while block
			   savedLoc2 = jump to end of while block */
			emitComment("BEGIN WHILE");
			savedLoc1 = emitSkip(0);
			emitComment("while: jump here after body comes back");
			/* generate code for condition */
			trav(tree->child[0]);
			/* Save one Line, to be filled in later */
			savedLoc2 = emitSkip(1);
			/* generate code for expression */
			trav(tree->child[1]);
			emitRM_Abs("LDA",pc,savedLoc1," while: jmp to end") ;
			/* Remember highest used number */
			currentLoc = emitSkip(0);
			/* Fill in saved Line for jump out of while block */
			emitBackup(savedLoc2);
			emitRM_Abs("JEQ",ac,currentLoc, "while: jmp out of while");

			emitRestore();
			emitComment("END WHILE");
			break;
		case KIND_ASSIGN:
			emitComment("-> assign");
			/* gencode for righthand */
			trav(tree->child[0]);
			loc = lookup(tree->name);
			if(tree->arr_offset > 0) {
				loc+=tree->arr_offset;
			}
			emitRM("ST",ac,loc,gp,"assign: store value");
			emitComment("<-assign");
			break;
		case KIND_READ:
			emitRO("IN",ac,0,0,"read integer value");
			loc = lookup(tree->child[0]->name);
			emitRM("ST",ac,loc,gp,"read:store value");
			break;
		case KIND_WRITE:
			/* gencode for expression */
			trav(tree->child[0]);
			emitRO("OUT",ac,0,0,"write ac");
			break;
		case KIND_OP:
			emitComment("-> OP");
			/* gencode for ac = lefthand */
			trav(tree->child[0]);
			/* gen code to push left operand */
			emitRM("ST",ac,tmpOffset--,mp,"op: push left");
			/* gencode for ac = righthand */
			trav(tree->child[1]);
			/* now load left operand from tmpOffset*/
			emitRM("LD",ac1,++tmpOffset,mp,"op: load left");
			switch (tree->op) {
				case OP_PLUS:
					emitRO("ADD",ac,ac1,ac,"op +");
					break;
				case OP_MINUS:
					emitRO("SUB",ac,ac1,ac,"op -");
					break;
				case OP_MUL:
					emitRO("MUL",ac,ac1,ac,"op *");
					break;
				case OP_DIV:
					emitRO("DIV",ac,ac1,ac,"op /");
					break;
				case OP_LS:
					emitRO("SUB",ac,ac1,ac,"op <");
					emitRM("JLT",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				case OP_GT:
					emitRO("SUB",ac,ac1,ac,"op >");
					emitRM("JGT",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				case OP_LE:
					emitRO("SUB",ac,ac1,ac,"op <");
					emitRM("JLE",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				case OP_GE:
					emitRO("SUB",ac,ac1,ac,"op >");
					emitRM("JGE",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				case OP_EQ:
					emitRO("SUB",ac,ac1,ac,"op ==");
					emitRM("JEQ",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				case OP_NE:
					emitRO("SUB",ac,ac1,ac,"op ==");
					emitRM("JNE",ac,2,pc,"br if true");
					emitRM("LDC",ac,0,ac,"false case");
					emitRM("LDA",pc,1,pc,"unconditional jmp");
					emitRM("LDC",ac,1,ac,"true case");
					break;
				default:
					emitComment("Bug: Unkown operator");
					break;
			}
			emitComment("<- OP");
			break;
		case KIND_CONST:
			emitComment("-> Const");
			emitRM("LDC",ac,tree->val,0,"load const");
			emitComment("<-Const");
			break;
		case KIND_ID:
			emitComment("-> ID");
			loc = lookup(tree->name);
			if (tree->arr_offset >0)
				loc+=tree->arr_offset;
			emitRM("LD",ac,loc,gp,"load id value");
			emitComment("<- ID");
			break;
		default:
			printnode(tree);
			emitComment("Bug: Unknown treenode");
			break;
	}

}


void trav(TreeNode *tree) {
	/* travers Syntax tree and calls gen */
	if (tree != NULL) {
		gen(tree);
		trav(tree->sibling);
	}
}

void codegen(TreeNode *tree, char *asmfilename){
	/* opens .vm file and writes header and footer */
	code = fopen(asmfilename,"w+");
	if (code == NULL) {
        	fprintf(stderr, "Could not open file %s\n", asmfilename);
        	exit(1);
       	}

	emitRM("LD",mp,0,ac,"load maxaddress from location 0");
   	emitRM("ST",ac,0,ac,"clear location 0");

	trav(tree);

	emitRO("HALT",0,0,0,"");
	fclose(code);
}
