%{
	/* $Id: parser.y,v 1.16 2006/01/17 21:10:45 arved Exp $ */

	#include <stdbool.h>
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>

	#include "globals.h"
	#include "parser.h"
	#include "symtab.h"
	#include "syntree.h"

	#ifdef	YYDEBUG
		extern int yydebug=1;
	#endif

	bool localvariable = false;

	bool erroroccured = false;
	int errorcount = 0;

%}
%start program
%union { /* SEMANTIC RECORD            */
char *id; 	/* For returning identifiers  */
int  num;      /* For returning numbers */
int type_int;
TreeNode *type_treenode;
}
%token KPROGRAM KWHILE KBOOL KPROC KREAD KTRUE KIF KWRITE KFALSE KELSE KINT
%token <num> NUMBER
%token OP_PLUS OP_LS OP_EQ OP_AS OP_MINUS OP_LE OP_NE OP_MUL OP_GT OP_DIV OP_GE
%token LPAREN RPAREN LBRACK RBRACK LBRACE RBRACE SC
%token <id> ID /* Simple identifier      */

%left OP_PLUS OP_MINUS
%left OP_MUL OP_DIV
%right KELSE

%type <id> subsetidentifier
%type <type_int> subsettype
%type <type_treenode> stmtlist number statement ifstmt whilestmt readstmt writestmt callstmt
%type <type_treenode> assignstmt expression var compoundstmt smplexp term factor
%%

program: 	/* leer */
		| KPROGRAM subsetidentifier LBRACE stmtlist RBRACE { savedTree = $4;}
		| KPROGRAM subsetidentifier LBRACE decllist proclist stmtlist RBRACE { savedTree = $6;}
		| KPROGRAM subsetidentifier LBRACE decllist stmtlist RBRACE { savedTree = $5;}
		| KPROGRAM subsetidentifier LBRACE proclist stmtlist RBRACE { savedTree = $5;}
		;
proclist:	procedure
		| proclist procedure
		;
procedure:	KPROC subsetidentifier LBRACE stmtlist RBRACE { declaration($2, TYPE_PROC, -1, $4);}
		| KPROC subsetidentifier LBRACE { localvariable = true; procname = strdup($2);} decllist stmtlist  { localvariable = false; free(procname); procname = NULL; } RBRACE
			{ declaration($2, TYPE_PROC, -1, $6);}
		;
decllist:	declaration
		| declaration decllist;
declaration:	subsettype subsetidentifier SC { declaration( $2 , $1, -1, NULL);}
		| subsettype subsetidentifier LBRACK number RBRACK SC
			{ checkarraydecl($4->val);
			  declaration($2, $1, $4->val, NULL);
			}
		;
subsettype: 	KINT  { $$ = TYPE_INT; }
		| KBOOL { $$ = TYPE_BOOL; }
		;
stmtlist:	stmtlist statement
               		{ TreeNode *t = $1;
                   		if (t != NULL) { /* NO Error or first statement*/
                   			while (t->sibling != NULL)
						/* traverse tree until last statement of stmtlist */
                        			t = t->sibling;
					/* Connect treenode of statment to tree of stmtlist */
                     			t->sibling = $2;
                     			$$ = $1;
				} else
					$$ = $2;
                 }

		| statement { $$ = $1; }
		;
statement:	ifstmt { $$ = $1; }
		| whilestmt { $$ = $1; }
		| readstmt { $$ = $1; }
		| writestmt { $$ = $1; }
		| callstmt { $$ = $1; }
		| assignstmt { $$ = $1; }
		| compoundstmt { $$ = $1; }
		| error SC { errorcount++; $$ = NULL;}
		;
ifstmt:		KIF LPAREN expression RPAREN statement %prec KELSE
			{ checkexpr($3);
			  $$ = createnodewchildren(KIND_IF, $3, $5,NULL);
			}
		|  KIF LPAREN expression RPAREN statement KELSE statement
			{ checkexpr($3);
			  $$ = createnodewchildren(KIND_IF, $3, $5, $7);
			}
		;
whilestmt:	KWHILE LPAREN expression RPAREN statement
			{ checkexpr($3);
			  $$ = createnodewchildren(KIND_WHILE, $3, $5, NULL);
			}
		;
readstmt:	KREAD var SC { $$ = createnodewchildren(KIND_READ, $2, NULL, NULL);}
		;
writestmt:	KWRITE expression SC
			{ checkwrite($2);
			  $$ = createnodewchildren(KIND_WRITE, $2, NULL, NULL);
			}
		;
var:		subsetidentifier LBRACK number RBRACK
			{ $$ = variable($1, $3->val);
			  checkarray($1,$3);
			}
		| subsetidentifier { $$ = variable($1, 0); }
		;
callstmt:	subsetidentifier LPAREN RPAREN SC { callstmt($1); $$ = symtablookupproctree($1);}
		;
assignstmt:	var OP_AS expression SC
			{ checkassignment($1, $3);
			  $$ = createnodewchildren(KIND_ASSIGN, $3, NULL, NULL);
			  $$->name = strdup($1->name);
			  $$->arr_offset = $1->arr_offset;
			}
		;
compoundstmt:	LBRACE stmtlist RBRACE { $$ = $2;}
		;
expression:	smplexp { $$ = $1; }
		| smplexp OP_LS smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_LS, TYPE_BOOL);
			}
		| smplexp OP_LE smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_LE, TYPE_BOOL);
			}
		| smplexp OP_GT smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_GT, TYPE_BOOL);
			}
		| smplexp OP_GE smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_GE, TYPE_BOOL);
			}
		| smplexp OP_EQ smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_EQ, TYPE_BOOL);
			}
		| smplexp OP_NE smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_NE, TYPE_BOOL);
			}
		;
smplexp:	term { $$ = $1; }
		| term OP_PLUS smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_PLUS, TYPE_INT);
			}
		| term OP_MINUS smplexp
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_MINUS, TYPE_INT);
			}
		;
term:		factor { $$ = $1; }
		| factor OP_MUL factor
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_MUL, TYPE_INT);
			}
		| factor OP_DIV factor
			{ checkoperand($1, $3);
			  $$ = createopnode(KIND_OP, $1, $3, NULL, OP_DIV, TYPE_INT);
			}
		;
factor:		LPAREN expression RPAREN { $$ = $2; }
		| number { $$ = $1; }
		| var { $$ = $1; }
		| KTRUE { $$ = newNode(KIND_CONST); $$->val = true; $$->type = TYPE_BOOL; }
		| KFALSE { $$ = newNode(KIND_CONST); $$->val = false; $$->type = TYPE_BOOL; }
		;

subsetidentifier:	ID { $$ = $1;}
			;

number:		NUMBER
			{ $$ = newNode(KIND_CONST);
			  $$->val = $1;
			  $$->type = TYPE_INT;
			}
		;
%%
	void yyerror(char *s)
	{
		fprintf(listingfilep, "\nline %d: %s\n", num_lines, s);
		fprintf(stderr, "\nline %d: %s\n", num_lines,  s);
		erroroccured= true;
	}

	void context_check( char *sym_name){
		if (procname) {
			if ( procsymtablookup(sym_name, 0,0,procname) == 1) {
				return;
			}
		}
		if ( symtablookup( sym_name,0,0 ) == -1 ) {
      			fprintf( stderr, "%s is an undeclared identifier\n", sym_name );
			yyerror("Undeclared identifier");
			erroroccured= true;
			errorcount++;
		}
 	}

	TreeNode *createnodewchildren(int kind, TreeNode *child1, TreeNode *child2, TreeNode *child3){
		TreeNode *t = newNode(kind);
		t->child[0] = child1;
		t->child[1] = child2;
		t->child[2] = child3;
		return t;
	}

	TreeNode *createopnode(int kind, TreeNode *child1, TreeNode *child2, TreeNode *child3, int op, int type){
		TreeNode *t = createnodewchildren(kind, child1, child2, child3);
		t->op = op;
		t->type = type;
		return t;
	}

	TreeNode *variable( char *sym_name, int arr_offset){
		context_check(sym_name);
		TreeNode *t = newNode(KIND_ID);
		if (procname != NULL) {
			t->name = createsymname(sym_name, procname);
		} else {
			t->name = strdup(sym_name);
		}
		t->type = symtablookuptype(t->name);
		t->arr_offset = arr_offset;
		return t;
	}

	void declaration(char *sym_name, int type, int arr_sz, TreeNode* proctree) {
		if (localvariable) {
			//fprintf(stderr,"Lokale Variable: %s\n", sym_name);

			if (procsymtablookup(sym_name, &type, arr_sz, procname) != -1) {
				fprintf( stderr, "%s is already declared\n", sym_name );
				yyerror("Variable already declared");
				erroroccured= true;
				errorcount++;
				return;
			}
			symtabinsert(sym_name, type, arr_sz, proctree, procname);
			//fprintf(stderr,"Lokale Variable named: %s\n", symnamenew);
		} else {
			if (symtablookup(sym_name, &type, arr_sz) != -1) {
				fprintf( stderr, "%s is already declared\n", sym_name );
				yyerror("Variable already declared");
				erroroccured= true;
				errorcount++;
				return;
			}
			symtabinsert(sym_name, type, arr_sz, proctree, procname);
		}
	}

	void callstmt(char *sym_name){
		context_check(sym_name);
		if (symtablookuptype(sym_name) != TYPE_PROC) {
			fprintf( stderr, "%s is an Invalid Function call\n", sym_name );
			yyerror("Invalid Function call");
			erroroccured= true;
			errorcount++;
		}
	}

	void checkarraydecl(int size) {
		if (size < 1) {
			fprintf( stderr, "%d is an Invalid Array size\n", size );
			yyerror("Invalid Array size");
			erroroccured= true;
			errorcount++;
		}
	}

	void checkarray(char *sym_name, TreeNode *arr_sz){
		int symarr_sz = symtablookuparr_sz(sym_name);
		if ((symarr_sz-1) < arr_sz->val) {
			fprintf( stderr, "%d is an Invalid Array subscript\n", arr_sz->val );
			yyerror("Invalid Array subscript");
			erroroccured= true;
			errorcount++;
		}
	}

	void checkoperand(TreeNode *node1, TreeNode *node2){
		if  ((node1->type != TYPE_INT) | (node2->type != TYPE_INT)) {
			fprintf(stderr, "Operand is not an integer: %d %d\n", node1->type ,node2->type );
			yyerror("Operand is not an integer\n");
			erroroccured= true;
			errorcount++;
		}
	}

	void checkassignment(TreeNode *node1, TreeNode *node2){
		if ( node1->type != node2->type) {
			if ( node2->type == TYPE_INT) {
				fprintf(stderr, "Assignment to non-integer\n");
				yyerror("Assignment to non-integer\n");
				erroroccured= true;
				errorcount++;
			}
			if ( node2->type == TYPE_BOOL) {
				fprintf(stderr, "Assignment to non-bool\n");
				yyerror("Assignment to non-bool\n");
				erroroccured= true;
				errorcount++;
			}
		}
	}

	void checkwrite(TreeNode *node) {
		if ((node->type != TYPE_INT) && (node->type != TYPE_BOOL)) {
			fprintf(stderr, "Write of non-integer, non-boolean\n");
			yyerror("Write of non-integer, non-boolean\n");
			erroroccured= true;
			errorcount++;
		}
	}

	void checkexpr(TreeNode *node) {
		if (node->type != TYPE_BOOL) {
			fprintf(stderr, "Condition not Boolean\n");
			yyerror("Condition not Boolean\n");
			erroroccured= true;
			errorcount++;
		}
	}

	#include "main.c"
