	/* $Id: scanner.l,v 1.7 2006/01/17 18:44:02 arved Exp $ */

	#include <stdbool.h>
	#include <string.h>
	#include "globals.h"
	#include "scanner.h"
	#include "y.tab.h"

	extern bool erroroccured;
	extern FILE *tokenfilep;
	extern FILE *listingfilep;
	extern int num_lines;

KPROGRAM	program
KWHILE		while
KBOOL		bool
KPROC		proc
KREAD		read
KTRUE		true
KIF		if
KWRITE		write
KFALSE		false
KELSE		else
KINT		int

ID		([a-zA-Z_][0-9a-zA-Z_]*)
NUMBER		[0-9]+

OP_PLUS		"+"
OP_LS		"<"
OP_EQ		"=="
OP_AS		"="
OP_MINUS	"-"
OP_LE		"<="
OP_NE		"!="
OP_MUL		"*"
OP_GT		">"
OP_DIV		"/"
OP_GE		">="

LPAREN		"("
RPAREN		")"
LBRACK		"["
RBRACK		"]"
LBRACE		"{"
RBRACE		"}"

SC		;

COMMENT_START	"/*"
COMMENT_END	"*/"

WHITESPACE	[ \t]+
NEWLINE		\n

%x comment
%%

"/*"         { BEGIN(comment); printlis("/*");}

<comment>[^*\n]*        {/* skip anything that's not a '*' */
			printlis(yytext);}
<comment>"*"+[^*/\n]*   {/* skip up '*'s not followed by '/'s */
			printlis(yytext);}
<comment>\n             printlisn(yytext);
<comment>"*"+"/"        { BEGIN(INITIAL); printlis("*/");}

{KPROGRAM}      { fprintf(tokenfilep, "KPROGRAM\n"); printlis(yytext); return(KPROGRAM);}
{KWHILE}	{ fprintf(tokenfilep, "KWHILE\n"); printlis(yytext); return(KWHILE);}
{KBOOL}		{ fprintf(tokenfilep, "KBOOL\n"); printlis(yytext); return(KBOOL);}
{KPROC}		{ fprintf(tokenfilep, "KPROC\n"); printlis(yytext); return(KPROC);}
{KREAD}		{ fprintf(tokenfilep, "KREAD\n"); printlis(yytext); return(KREAD);}
{KTRUE}		{ fprintf(tokenfilep, "KTRUE\n"); printlis(yytext); return(KTRUE);}
{KIF}		{ fprintf(tokenfilep, "KIF\n"); printlis(yytext); return(KIF);}
{KWRITE}	{ fprintf(tokenfilep, "KWRITE\n"); printlis(yytext); return(KWRITE);}
{KFALSE}	{ fprintf(tokenfilep, "KFALSE\n"); printlis(yytext); return(KFALSE);}
{KELSE}		{ fprintf(tokenfilep, "KELSE\n"); printlis(yytext); return(KELSE);}
{KINT}		{ fprintf(tokenfilep, "KINT\n");
		printlis(yytext);
		return(KINT);
		}
{OP_PLUS}	{ fprintf(tokenfilep, "OP_PLUS\n"); printlis(yytext); return(OP_PLUS);}
{OP_MINUS}	{ fprintf(tokenfilep, "OP_MINUS\n"); printlis(yytext); return(OP_MINUS);}
{OP_MUL}	{ fprintf(tokenfilep, "OP_MUL\n"); printlis(yytext); return(OP_MUL);}
{OP_DIV}	{ fprintf(tokenfilep, "OP_DIV\n"); printlis(yytext); return(OP_DIV);}
{OP_LS}		{ fprintf(tokenfilep, "OP_LS\n"); printlis(yytext); return(OP_LS);}
{OP_LE}		{ fprintf(tokenfilep, "OP_LE\n"); printlis(yytext); return(OP_LE);}
{OP_GT}		{ fprintf(tokenfilep, "OP_GT\n"); printlis(yytext); return(OP_GT);}
{OP_GE}		{ fprintf(tokenfilep, "OP_GE\n"); printlis(yytext); return(OP_GE);}
{OP_EQ}		{ fprintf(tokenfilep, "OP_EQ\n"); printlis(yytext); return(OP_EQ);}
{OP_NE}		{ fprintf(tokenfilep, "OP_NE\n"); printlis(yytext); return(OP_NE);}
{OP_AS}		{ fprintf(tokenfilep, "OP_AS\n"); printlis(yytext); return(OP_AS);}
{LPAREN}	{ fprintf(tokenfilep, "LPAREN\n"); printlis(yytext); return(LPAREN);}
{RPAREN}	{ fprintf(tokenfilep, "RPAREN\n"); printlis(yytext); return(RPAREN);}
{LBRACK}	{ fprintf(tokenfilep, "LBRACK\n"); printlis(yytext); return(LBRACK);}
{RBRACK}	{ fprintf(tokenfilep, "RBRACK\n"); printlis(yytext); return(RBRACK);}
{LBRACE}	{ fprintf(tokenfilep, "LBRACE\n"); printlis(yytext); return(LBRACE);}
{RBRACE}	{ fprintf(tokenfilep, "RBRACE\n"); printlis(yytext); return(RBRACE);}
{SC}		{ fprintf(tokenfilep, "SC\n"); printlis(yytext); return(SC);}

{NUMBER}	{fprintf(tokenfilep, "NUMBER %s\n", yytext); yylval.num = strtol(yytext, (char **)NULL, 10); printlis(yytext); return(NUMBER);}
{ID}  		{ fprintf(tokenfilep, "ID %s\n", yytext);
		printlis(yytext);
		yylval.id = (char *) strdup(yytext);
		return(ID);}
{WHITESPACE}	{ printlis(yytext);}
{NEWLINE}	{ printlisn(yytext); }
.		{ fprintf(listingfilep, "\n    INVALID SYMBOL '%s'\n", yytext);
		fprintf(stdout, "\nline %d: INVALID SYMBOL '%s'\n",num_lines, yytext);
		erroroccured = true;
		}
%%
	void
	printlis(const char* text)
	{
		if (firsttoken) {
			fprintf(listingfilep, "%03d ", num_lines);
			firsttoken = false;
		}
		fprintf(listingfilep, text);
		}

	void
	printlisn(const char* text)
	{
		printlis(text);
		num_lines++;
		fprintf(listingfilep, "%03d ", num_lines);
	}

