/* $Id: symtab.c,v 1.8 2006/01/17 21:03:52 arved Exp $ */

#ifdef __FreeBSD__
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "globals.h"
/* Counter for variable memory adresses */
int location = 0;

SLIST_HEAD(slisthead, entry) head = SLIST_HEAD_INITIALIZER(head);
struct slisthead *headp;                /* Singly-linked List head. */
struct entry {
        char *name;
        int type;
        int arr_sz;
        int addr;
        TreeNode *proctree;
        SLIST_ENTRY(entry) entries;
} *entryp;

char *createsymname(char *mysym_name, char *myprocname) {
	char *symnamenew = malloc(( strlen(mysym_name) +\
		strlen(myprocname) +2) * sizeof(char) );
	memset(symnamenew, 0 , sizeof(symnamenew));
	symnamenew = strcat(symnamenew, myprocname);
	symnamenew = strcat(symnamenew, "-");
	symnamenew = strcat(symnamenew, mysym_name);
	return symnamenew;
}

void symtabinsert(char *sym_name, int type, int arr_sz, TreeNode *proctree, char *entryprocname) {
	entryp = malloc(sizeof(struct entry));
	if (entryprocname) {
		char *symnamenew = createsymname(sym_name, entryprocname);
		entryp->name = strdup(symnamenew);
	} else {
		entryp->name = strdup(sym_name);
	}
	entryp->type = type;
	entryp->arr_sz = arr_sz;
	entryp->proctree = proctree;
	entryp->addr = location;
	//fprintf(stderr, "Insert: %s at %d\n", sym_name, location);
	if (arr_sz != -1)
		location += arr_sz;
	else
		location++;
	SLIST_INSERT_HEAD(&head, entryp, entries);
}

int symtablookup(char *sym_name, int *type, int arr_sz){
	SLIST_FOREACH(entryp, &head, entries) {
		if (strcmp (entryp->name,sym_name) == 0) {
			return 1;
		}
	}
	return -1;
}

int procsymtablookup(char *sym_name, int *type, int arr_sz, char *entryprocname){
	char *symnamenew = createsymname(sym_name,entryprocname);

	return symtablookup(symnamenew, type, arr_sz);
}

int symtablookuptype(char *sym_name){
	SLIST_FOREACH(entryp, &head, entries) {
		if (strcmp (entryp->name,sym_name) == 0)
			return entryp->type;
	}
	return -1;
}

int symtablookuparr_sz(char *sym_name){
	SLIST_FOREACH(entryp, &head, entries) {
		if (strcmp (entryp->name,sym_name) == 0)
			return entryp->arr_sz;
	}
	return -1;
}

int lookup(char *sym_name) {
	SLIST_FOREACH(entryp, &head, entries) {
		if (sym_name) {
			if (strcmp (entryp->name,sym_name) == 0) {
				return entryp->addr;
			}
		} else {
			fprintf(stderr,"lookup: sym_name empty!\n");
			return -1;
		}
	}
	fprintf(stderr,"lookup: Variable not found!\n");
	return -1; /* nothing found */
}

TreeNode *symtablookupproctree(char *sym_name){
		SLIST_FOREACH(entryp, &head, entries) {
		if (strcmp (entryp->name,sym_name) == 0)
			return entryp->proctree;
	}
	return NULL;

}

void symtabfree() {
	while(!SLIST_EMPTY(&head)) {
		entryp = SLIST_FIRST(&head);
		SLIST_REMOVE_HEAD(&head, entries);
		free(entryp);
	}
}

void symtabinit() {
	SLIST_INIT(&head);
}
